--[[
######################################################################
##  File:	AirBallon3.lua											##
##  Author:	'OliveD' (olived@live.ca)								##
##	Notes:	Script for the air balloon passenger and movement.		##
##			Rewritten on Asklar's request. Now supports passenger	##
##			inventories, return to orbit and other such.			##
######################################################################
]]--
function FixActors(self)
	local i;
	for i=0,self.InventorySize-1,1 do
		local myIterator = self.Inventory();
		if myIterator.Team == self.Team and myIterator.ClassName == "AHuman" and myIterator.Sharpness==94 then
			local myFixSpawner = CreateAHuman("Militia 1");
			myFixSpawner.Team = self.Team;
			MovableMan:AddActor(myFixSpawner);
			--self:SwapNextInventory(myFixSpawner,true);
			--myFixSpawner.ToDelete = true;
		else
			self:SwapNextInventory(myIterator,true);
		end
	end
end

function ApplyHumanProperties(target,source)
	target.Health = source.Health;
	if source.FGArm == nil then
		target.FGArm:GibThis();
	end
	if source.BGArm == nil then
		target.BGArm:GibThis();
	end
	if source.FGLeg == nil then
		target.FGLeg:GibThis();
	end
	if source.BGLeg == nil then
		target.BGLeg:GibThis();
	end
	for i=0,source.InventorySize-1,1 do
		local Iteratorb = source.Inventory();
		if Iteratorb.ClassName == "HDFirearm" then
			target:AddInventoryItem(CreateHDFirearm(Iteratorb.PresetName));
		elseif Iteratorb.ClassName == "HeldDevice" then
			target:AddInventoryItem(CreateHeldDevice(Iteratorb.PresetName));
		elseif Iteratorb.ClassName == "TDExplosive" then
			target:AddInventoryItem(CreateTDExplosive(Iteratorb.PresetName));
		end
		source:SwapNextInventory(myIteratorb,true);
	end
end
function ApplyCrabProperties(target,source)
	target.Health = source.Health;
	if source.Turret == nil then
		self.myCrab.Turret:GibThis();
	end
	if source.LeftFGLeg == nil then
		self.myCrab.LeftFGLeg:GibThis();
	end
	if source.LeftBGLeg == nil then
		self.myCrab.LeftBGLeg:GibThis();
	end
	if source.RightFGLeg == nil then
		self.myCrab.RightFGLeg:GibThis();
	end
	if source.RightBGLeg == nil then
		self.myCrab.RightBGLeg:GibThis();
	end
end

function BasketAdd(self)
	local i;
	for i=0,self.InventorySize-1,1 do
		local myIterator = self.Inventory();
		if myIterator.Team == self.Team and myIterator.ClassName == "AHuman" then
			self.myBasketActor = CreateAHuman(myIterator.PresetName);
			--This is saved in the object because we need to apply health and limb wounds to it later.
			self.myHuman = ToAHuman(myIterator); 
			--equipped item
			if ToAHuman(myIterator).EquippedItem ~= nil then
				self.myBasketActor:AddInventoryItem(CreateHDFirearm(self.myHuman.EquippedItem.PresetName));
			end
			--do properties
			ApplyHumanProperties(self.myBasketActor,self.myHuman);
			self.myBasketActor.Team = self.Team;
			MovableMan:AddActor(self.myBasketActor);
			break;
		elseif myIterator.Team == self.Team and myIterator.ClassName == "ACrab" then
			--This is saved in the object because we need to apply health and limb wounds to it later.
			self.myCrab = ToACrab(myIterator); 
			self.myBasketActor = CreateACrab(myIterator.PresetName);
			ApplyCrabProperties(self.myBasketActor,self.myHuman);
			self.myBasketActor.Team = self.Team;
			MovableMan:AddActor(self.myBasketActor);
			break;
		end
		self:SwapNextInventory(myIterator,true);
	end
end

function BasketUpdate(self)
	if self.myBasketActor ~= nil then
		self.myBasketActor.Pos = self.Pos + Vector(self.RotAngle + self.myBasketMO.RotAngle,40);
		self.myBasketActor.Vel = self.Vel;
		self.myBasketActor.AngularVel = 0;
		self.myBasketActor.RotAngle = 0;
		--That would be bad.
		self.myBasketActor:GetController():SetState(Controller.MOVE_UP,false);
	end
end

function BasketCleanUp(self)
	if self.myBasketActor ~= nil then
		if self.myHuman ~= nil then
			ApplyHumanProperties(self.myHuman,self.myBasketActor);
		elseif self.myCrab ~= nil then
			ApplyCrabProperties(self.myCrab,self.myBasketActor);
		end
		self.myBasketActor.ToDelete = true;
		self.myBasketActor = nil;
		self.myHuman = nil;
		self.myCrab = nil;
	end
end

function Create(self)
	self.flipTimer = Timer();
	self.IgnoresTeamHits = true;
	self.myBasketMO = nil;
	self.myBasketActor = nil;
	--O basket where art thou
	local i;
	for i=1,MovableMan:GetMOIDCount()-1,1 do
		if MovableMan:GetMOFromID(i).RootID == self.ID and i ~= self.ID then
			local iterator = MovableMan:GetMOFromID(i);
			if string.find(iterator.PresetName, "Loyalists Air Balloon Basket") then
				self.myBasketMO = ToAttachable(iterator);
			end
		end
		i = i + 1
	end
	if self.myBasketMO == nil then
		print("Loyalists Air Balloon: Could not find basket on Create(). MOID limit error maybe?");
		self:GibThis();
	end
	FixActors(self);
end

function Update(self)
	if self:GetController():IsState(Controller.MOVE_LEFT) == false and self:GetController():IsState(Controller.MOVE_RIGHT) == false then
		self.Vel.X = self.Vel.X*.96;
		self.RotAngle = self.RotAngle*.96;
	end
	if self.myBasketMO ~= nil then
		if self.myBasketMO:IsAttached() then
			self.myBasketMO.AngleRot = math.rad(self.Vel.X*.6);
			if self.HatchState == 0 then
				--basket actor spawning stuff
				if self.myBasketActor == nil then
					if self.Pos.Y >= 5 then
						BasketAdd(self);
					end
				else
					if self.Pos.Y <	5 then
						BasketCleanUp(self);
					end
					BasketUpdate(self);
				end
			else
				--clean up the basket actor if we have one since the doors are open atm.
				BasketCleanUp(self);
			end
		end
	end
end

function Destroy(self)
	--clean up the basket actor
	BasketCleanUp(self);
end